import os
import sys
import subprocess
import re
import time
from datetime import datetime
try:
    from set_gpu import set_gpu
except:
    from .set_gpu import set_gpu


def benchmark():
    print('显卡(cuda:0)性能测试: cuda 11.7 下 3090Ti=85(理论160), V100 16GB=86(理论125), A100 80GB=231(理论312); cuda 11.4 下 2080Ti=58')
    import torch
    from torch.utils import benchmark
    typ = torch.float16
    n = 1024 * 16
    a = torch.randn(n, n).type(typ).cuda()
    b = torch.randn(n, n).type(typ).cuda()
    t = benchmark.Timer(stmt='a @ b', globals={'a': a, 'b': b})
    x = t.timeit(50)
    print(2*n**3 / x.median / 1e12)


def wait_gpus(para_L):
    args = {
        '-c': 0.1,  # 至少剩余多少gpu使用率, 百分比
        '-m': 0.8,  # 至少剩余多少显存, 小于1是百分比, 大于1是MB
        '-t': 10,  # 这些显卡至少连续几秒钟满足这些条件
    }
    i = 0
    for p in para_L:
        if '=' in p and p.split('=')[0] in args:
            args[p.split('=')[0]] = float(p.split('=')[1])
        else:
            break
        i += 1
    para_L = para_L[i:]
    # 获取gpus
    gpus = []
    cmd = ' ' + ' '.join(para_L) + ' '
    gpus_re = [
        r'(?<= --include=localhost:)\d[\d,]*(?= )',  # deepspeed
        r'(?<= CUDA_VISIBLE_DEVICES=)\d[\d,]*(?= )',
    ]
    for r in gpus_re:
        ret = re.search(r, cmd)
        if ret:
            gpus = [int(i) for i in ret.group().strip(',').split(',')]
            break
    print(cmd.strip())
    print('等待GPU {} 满足条件:'.format(gpus), args, datetime.now())
    gpu_info = {i: None for i in gpus}  # {序号:满足要求的时间,..}
    start = time.time()
    while True:
        ret = set_gpu(return_more=True)
        for g in gpu_info.keys():
            gpu_usage = ret['gpu_usage'][g]
            ext_gpu_mem = ret['ext_gpu_mem'][g]
            all_gpu_mem = ret['all_gpu_mem'][g]
            mem_rate = args['-m'] / all_gpu_mem if args['-m'] > 1 else args['-m']
            if (100 - gpu_usage) / 100 >= args['-c'] and ext_gpu_mem / all_gpu_mem >= mem_rate:
                gpu_info[g] = time.time() if gpu_info[g] is None else gpu_info[g]
            else:
                gpu_info[g] = None
        bypass = False if gpu_info else True
        for t in gpu_info.values():
            if t is not None and time.time() - t > args['-t']:
                bypass = True
            else:
                bypass = False
                break
        if bypass:
            print('等待 {} 秒GPU满足条件开始运行'.format(round(time.time()-start,2)), datetime.now())
            os.system(cmd.strip())
            break
        time.sleep(1)


def main():
    para = sys.argv[1:]
    if para:
        if para[0] == '--wait':  # 一个等待gpu的功能
            wait_gpus(para[1:])
            return
        elif para[0] == '--benchmark':  # 测试显卡性能
            benchmark()
            return
    # 一些转义符号复原
    for i, p in enumerate(para):
        if '\\' in p:
            para[i] = para[i].replace('\\', '\\\\')
        if '"' in p:
            para[i] = '"' + para[i].replace('"', '\\"') + '"'
        if '$' in p:
            para[i] = para[i].replace('$', '\\\\\\$')  # 因为 shell eval 所以多加入转义
    # 自动寻找 auto.sh 的位置
    auto = subprocess.getstatusoutput('py=$(which python) && echo ${py%bin*}lib/python*/site-packages/tsc_auto/auto.sh')
    # 没有报错
    if auto[0] == 0:
        auto = auto[1]
    else:
        raise NameError(str(auto) + ' 寻找auto.sh错误!')
    # 运行
    cmd = f'chmod 777 {auto} && {auto} ' + ' '.join(para)
    os.system(cmd)


if __name__ == '__main__':
    main()



import requests
import json
try:
    from set_gpu import set_gpu
except:
    from .set_gpu import set_gpu
from datetime import datetime
import argparse
import sys
import os
import fcntl
import time
import collections


def send_wechat(title: str, content: str, token: str, template='txt', channel='wechat'):
    """发送消息接口: http://www.pushplus.plus/doc/guide/api.html

    Args:
        title (str): 消息标题
        content (str): 具体消息内容，根据不同template支持不同格式
        token (str): 用户令牌
        template (str, optional): 发送模板, 例如 txt html json markdown
        channel (str, optional): 发送渠道

    Returns:
        ret (dict)
    """
    data = {
        'token': token,
        'title': title,
        'content': content,
        'template': template,
        'channel': channel,
    }
    ret = {"msg": "无法联网, 通知失败!"}
    for i in range(3):  # 失败后再尝试
        try:
            r = requests.post(url='https://www.pushplus.plus/send', data=json.dumps(data))
            ret = json.loads(r.text)
            break
        except:
            print('notification failed {} ...'.format(i+1))
            time.sleep(5)
    return ret


def get_time_diff(s: str, e: str):
    """自动计算两个时间的差, 自动赋予单位, 秒/分钟/小时/天

    Args:
        s (str): 开始时间, 格式: date "+%Y-%m-%d %H:%M:%S.%N|cut -c 1-26" = 2022-05-24 18:18:59.574227
        e (str): 结束时间, 格式和s一样

    Returns:
        str: 例如 3.1 秒
    """
    try:
        s = datetime.strptime(s, "%Y-%m-%d %H:%M:%S.%f")
        e = datetime.strptime(e, "%Y-%m-%d %H:%M:%S.%f")
    except:
        return ''
    diff = datetime.timestamp(e) - datetime.timestamp(s)
    unit = '秒'
    if diff > 120:
        diff = diff / 60
        unit = '分钟'
        if diff > 120:
            diff = diff / 60
            unit = '小时'
            if diff > 48:
                diff = diff / 24
                unit = '天'
    return '{} {}'.format(round(diff, 2), unit)


def explain_exit_code(code):
    """返回linux错误代码的一般含义和提示
    参考: 
        https://tldp.org/LDP/abs/html/exitcodes.html
        https://www.computerhope.com/unix/signals.htm
        https://blog.csdn.net/weixin_34261739/article/details/91785562

    Args:
        code (int, float, str): 错误代码

    Returns:
        dict: {'code':code,'meaning':'含义','comment':'提示'}; 没有一般含义的代码和可靠信号SIGRTMIN/SIGRTMAX等只有code
    """
    code_status = {
        0: {'meaning': '正常退出', 'comment': '程序执行完成'},
        1: {'meaning': '一般错误', 'comment': '例如“除以零”和其他禁止操作: let "var1 = 1/0"'},
        2: {'meaning': '错误使用 shell 内置程序(根据 Bash 文档)', 'comment': '缺少关键字或命令，或权限问题（如diff在二进制文件比较失败时的返回代码）'},
        126: {'meaning': '调用的命令无法执行', 'comment': '权限问题或命令不是可执行文件'},
        127: {'meaning': '“找不到相关命令”', 'comment': '例如 $PATH 或命令拼写错误'},
        128: {'meaning': '退出代码无效', 'comment': 'exit 不接受[0,255]整数以外的数: exit 3.14159'},
        128+1: {'meaning': '收到中断信号(1)SIGHUP', 'comment': '该信号通常表示控制终端或虚拟终端已关闭'},
        128+2: {'meaning': '收到中断信号(2)SIGINT', 'comment': '程序终止(interrupt)信号, 在用户键入INTR字符(通常是Ctrl-C)时发出，用于通知前台进程组终止进程'},
        128+3: {'meaning': '收到中断信号(3)SIGQUIT', 'comment': '和SIGINT类似, 但由QUIT字符(通常是Ctrl-\)来控制. 进程在因收到SIGQUIT退出时会产生core文件, 在这个意义上类似于一个程序错误信号'},
        128+4: {'meaning': '收到中断信号(4)SIGILL', 'comment': '执行了非法指令. 通常是因为可执行文件本身出现错误, 或者试图执行数据段. 堆栈溢出时也有可能产生这个信号'},
        128+5: {'meaning': '收到中断信号(5)SIGTRAP', 'comment': '由断点指令或其它trap指令产生. 由debugger使用'},
        128+6: {'meaning': '收到中断信号(6)SIGABRT', 'comment': '调用abort函数生成的信号'},
        128+7: {'meaning': '收到中断信号(7)SIGBUS', 'comment': '非法地址, 包括内存地址对齐(alignment)出错。比如访问一个四个字长的整数, 但其地址不是4的倍数'},
        128+8: {'meaning': '收到中断信号(8)SIGFPE', 'comment': '在发生致命的算术运算错误时发出. 不仅包括浮点运算错误, 还包括溢出及除数为0等其它所有的算术的错误'},
        128+9: {'meaning': '收到中断信号(9)SIGKILL', 'comment': '用来立即结束程序的运行, 例如内存不足被系统终止'},
        128+10: {'meaning': '收到中断信号(10)SIGUSR1', 'comment': '留给用户使用'},
        128+11: {'meaning': '收到中断信号(11)SIGSEGV', 'comment': '试图访问未分配给自己的内存, 或试图往没有写权限的内存地址写数据'},
        128+12: {'meaning': '收到中断信号(12)SIGUSR2', 'comment': '留给用户使用'},
        128+13: {'meaning': '收到中断信号(13)SIGPIPE', 'comment': '管道破裂。这个信号通常在进程间通信产生，比如采用FIFO(管道)通信的两个进程，读管道没打开或者意外终止就往管道写，写进程会收到SIGPIPE信号。此外用Socket通信的两个进程，写进程在写Socket的时候，读进程已经终止'},
        128+14: {'meaning': '收到中断信号(14)SIGALRM', 'comment': '时钟定时信号, 计算的是实际的时间或时钟时间. alarm函数使用该信号'},
        128+15: {'meaning': '收到中断信号(15)SIGTERM', 'comment': '程序结束(terminate)信号, 与SIGKILL不同的是该信号可以被阻塞和处理。通常用来要求程序自己正常退出'},
        128+16: {'meaning': '收到中断信号(16)SIGSTKFLT', 'comment': '堆栈故障。映射到Linux中的SIGUNUSED'},
        128+17: {'meaning': '收到中断信号(17)SIGCHLD', 'comment': '子进程结束时, 父进程会收到这个信号'},
        128+18: {'meaning': '收到中断信号(18)SIGCONT', 'comment': '让一个停止(stopped)的进程继续执行'},
        128+19: {'meaning': '收到中断信号(19)SIGSTOP', 'comment': '停止(stopped)进程的执行'},
        128+20: {'meaning': '收到中断信号(20)SIGTSTP', 'comment': '停止进程的运行, 但该信号可以被处理和忽略. 用户键入SUSP字符时(通常是Ctrl-Z)发出这个信号'},
        128+21: {'meaning': '收到中断信号(21)SIGTTIN', 'comment': '当后台作业要从用户终端读数据时, 该作业中的所有进程会收到SIGTTIN信号. 缺省时这些进程会停止执行'},
        128+22: {'meaning': '收到中断信号(22)SIGTTOU', 'comment': '类似于SIGTTIN, 但在写终端(或修改终端模式)时收到'},
        128+23: {'meaning': '收到中断信号(23)SIGURG', 'comment': '有"紧急"数据或out-of-band数据到达socket时产生'},
        128+24: {'meaning': '收到中断信号(24)SIGXCPU', 'comment': '超过CPU时间资源限制. 这个限制可以由getrlimit/setrlimit来读取/改变'},
        128+25: {'meaning': '收到中断信号(25)SIGXFSZ', 'comment': '当进程企图扩大文件以至于超过文件大小资源限制'},
        128+26: {'meaning': '收到中断信号(26)SIGVTALRM', 'comment': '虚拟时钟信号. 类似于SIGALRM, 但是计算的是该进程占用的CPU时间'},
        128+27: {'meaning': '收到中断信号(27)SIGPROF', 'comment': '类似于SIGALRM/SIGVTALRM, 但包括该进程用的CPU时间以及系统调用的时间'},
        128+28: {'meaning': '收到中断信号(28)SIGWINCH', 'comment': '窗口大小改变时发出'},
        128+29: {'meaning': '收到中断信号(29)SIGIO,SIGPOLL', 'comment': '文件描述符准备就绪, 可以开始进行输入/输出操作'},
        128+30: {'meaning': '收到中断信号(30)SIGPWR,SIGLOST', 'comment': '系统检测到停电时'},
        128+31: {'meaning': '收到中断信号(31)SIGUNUSED,SIGSYS', 'comment': '非法的系统调用. 此信号是出于兼容性原因提供的，例如在Linux中从具有不同或不支持信号的操作系统移植软件时'},
        255: {'meaning': '退出状态超出范围', 'comment': 'exit 不接受[0,255]整数以外的数: exit -1'},
    }
    try:
        code = int(code)
        status = {'code': code}
        if code in code_status:
            status.update(code_status[code])
    except:
        status = {'code': code}
    return status


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='输入参数')
    parser.add_argument('-t', default='', help='token')
    parser.add_argument('-c', default='', help='执行指令')
    parser.add_argument('-d', default='', help='命令执行的退出代码')
    parser.add_argument('-s', default='', help='程序开始执行时间, 格式参考 get_time_diff')
    parser.add_argument('-e', default='', help='程序结束执行时间, 格式参考 get_time_diff')
    args, unknown = parser.parse_known_args()  # 忽略未知参数
    # print(args, unknown)
    if args.c == '':
        fcntl.fcntl(sys.stdin, fcntl.F_SETFL, os.O_NONBLOCK)
        try:
            cc = sys.stdin.read()
            args.c = cc.strip()
        except:
            ...
    if args.c == '' or args.t == '':
        sys.exit(0)
    print('结束执行时间:', args.e, ', 总耗时:', get_time_diff(args.s, args.e), ', 退出代码:', args.d)
    machine = set_gpu(return_more=True, public_net=True)
    content = {
        '总耗时': get_time_diff(args.s, args.e),
        '执行目录': os.getcwd(),
        '开始执行时间': args.s,
        '结束执行时间': args.e,
        '执行指令': args.c,
        '命令退出状态(UNIX信号解释)': explain_exit_code(args.d),
        '执行结束后的机器状态(mem单位为MB)': {k: '; '.join([str(i) for i in v]) if isinstance(v, list) else v for k, v in machine.items()},
        '当前环境变量': collections.OrderedDict(sorted(os.environ.items(), key=lambda t: t[0])),
    }
    title = '{} {}: {}'.format(machine['hostname'], args.d, args.c)[:100]
    ret = send_wechat(title, json.dumps(content, ensure_ascii=False, indent=4)[:10000], args.t)
    print('notification:', ret)



# 自动杀死高资源占用进程
import subprocess
import re
import time
import sys
import os
import traceback
import copy
from datetime import datetime
from pprint import pprint
import argparse
import shutil


def get_nvidia_processes():
    (status, result) = subprocess.getstatusoutput('nvidia-smi')
    if status == 0:
        result = [re.split('[\s]+', i)[1:] for i in re.findall(r'[^\n\r]+(?=MiB)', result.split('Processes: ')[1])]
        result_L = []  # [{},..]
        for r in result:
            result_L.append({
                'gpu': r[0],  # 编号, 即 CUDA_VISIBLE_DEVICES
                'pid': r[-4],  # 进程PID
                'type': r[-3],  # 程序类型, 比如C
                'process': r[-2],  # 进程名
                'mem': float(r[-1]),  # 显存使用量(MiB)
            })
        return result_L
    return []


def get_ps_processes():
    means = [  # [(描述,处理函数),..]; 与返回结果每列依次对应
        ('user', str),  # 用户
        ('pid', str),  # 进程PID
        ('stat', str),  # 程序状态
        ('cpu', float),  # CPU使用率, 100相当于一个超线程占满
        ('mem', lambda x: int(x) / 1024),  # 内存(MB)
        ('vsz', lambda x: int(x) / 1024),  # 虚拟内存(MB)
        ('etime', lambda x: int(x) / 60),  # 进程运行时间(分钟)
        ('ctime', str),  # CPU累计时间([[DD-]hh:]mm:ss)
        ('command', str),  # 命令
    ]
    command = 'ps axo "user:100" -o " | %p | " -o "stat" -o " | %C | " -o "size" -o " | %z | " -o "etimes" -o " | %t | %a" --sort pid --width 10000'
    # 输出一行例如: root | 37170 | S |  0.0 |   152 |   1512 | 1210445 | 14-00:14:05 | tail -f /running
    (status, result) = subprocess.getstatusoutput(command)
    if status != 0:
        return []
    result_L = []  # [{},..]
    for line in result.split('\n')[1:]:
        line = [i.strip() for i in re.split(r' [|] ', line)]
        result_L.append({})
        for i, (j, k) in enumerate(means):
            result_L[-1][j] = k(line[i])
    return result_L


def get_user_processes():
    pid_gpu8mem_D = {}  # {pid:{gpu:memory,..},..}
    ps_processes = get_ps_processes()
    for result in get_nvidia_processes():
        pid = result['pid']
        gpu = result['gpu']
        memory = result['mem']
        if pid in pid_gpu8mem_D:
            if gpu in pid_gpu8mem_D[pid]:
                pid_gpu8mem_D[pid][gpu] += memory
            else:
                pid_gpu8mem_D[pid][gpu] = memory
        else:
            pid_gpu8mem_D[pid] = {gpu: memory}
    user_processes = {}
    for result in ps_processes:
        user_process = user_processes.setdefault(result['user'], {  # 每个用户记录内容
            'cpu': 0,  # 总cpu使用率
            'mem': 0,  # 总内存GB
            'vsz': 0,  # 总虚拟内存GB
            'gmem': 0,  # 总显存GB
            'gtime': 0,  # 总gpu运行小时
            'gcard': set(),  # 所有占用的显卡编号
            'gpid': set(),  # 所有占用显卡的pid
            'pro': [],  # [{},..]; 每个进程信息和{gpu:memory,..}, 参见 get_ps_processes()
        })
        user_process['cpu'] += result['cpu']
        user_process['mem'] += result['mem'] / 1024
        user_process['vsz'] += result['vsz'] / 1024
        del result['user']
        user_process['pro'].append(result)
        if result['pid'] in pid_gpu8mem_D:
            user_process['gmem'] += sum(pid_gpu8mem_D[result['pid']].values()) / 1024
            user_process['gtime'] += result['etime'] / 60
            user_process['gcard'] |= set(pid_gpu8mem_D[result['pid']])
            user_process['gpid'].add(result['pid'])
            user_process['pro'][-1]['gpu_mem'] = pid_gpu8mem_D[result['pid']]
    return user_processes


def kill_processes(config_, test=False):
    user_processes = get_user_processes()
    for i in set(config_['ignore_u']) - set(config_['include_u']):
        if i in user_processes:
            del user_processes[i]
    if test:
        print('user_processes:')
        pprint(user_processes)
    for user, v in user_processes.items():
        config = config_.copy()
        # 特殊配置的用户
        for conf in config_['conf']:
            if user in conf['conf_u']:
                config.update(conf)
        # 用户的特殊设置
        if user in config_['user']:
            config.update(config_['user'][user])
        config['user'] = None
        config['conf'] = None
        config['conf_u'] = None
        if test:
            print('使用的 config ：')
            pprint(config)
        #
        del_pids = {}  # {pid:{},..}; 需要杀死的pid和进程
        error_pids = {}  # {'错误信息':{要删除的pid,..}}
        processes = sorted(v['pro'], key=lambda t: t['etime'])  # 运行时间顺序, 优先杀死最新运行的程序
        processes_gpu = [i for i in processes if i['pid'] in v['gpid']]  # gpu 程序
        #
        x, n = v['cpu'], 0
        error = '用户占用的cpu百分比 = ' + str(x) + ' > ' + str(config['cpu_core_u'])
        while x > config['cpu_core_u']:
            p = processes[n]  # 取出一个进程
            del_pids[p['pid']] = p  # 加入删除
            x -= p['cpu']
            n += 1
            error_pids.setdefault(error, set())  # 错误信息
            error_pids[error].add(p['pid'])  # 加入错误pid
        #
        x, n = v['mem'], 0
        error = '用户占用的内存(GB) = ' + str(x) + ' > ' + str(config['cpu_mem_u'])
        while x > config['cpu_mem_u']:
            p = processes[n]  # 取出一个进程
            del_pids[p['pid']] = p  # 加入删除
            x -= p['mem'] / 1024
            n += 1
            error_pids.setdefault(error, set())  # 错误信息
            error_pids[error].add(p['pid'])  # 加入错误pid
        #
        x, n = len(v['gcard']), 0
        error = '用户占用的显卡数量 = ' + str(x) + ' > ' + str(config['gpu_card_u'])
        while x > config['gpu_card_u']:
            p = processes_gpu[n]  # 取出一个进程
            del_pids[p['pid']] = p  # 加入删除
            x = len(set(j for i in processes_gpu[n + 1:] for j in i['gpu_mem'].keys()))
            n += 1
            error_pids.setdefault(error, set())  # 错误信息
            error_pids[error].add(p['pid'])  # 加入错误pid
        #
        x, n = v['gmem'], 0
        error = '用户占用的显存(GB) = ' + str(x) + ' > ' + str(config['gpu_mem_u'])
        while x > config['gpu_mem_u']:
            p = processes_gpu[n]  # 取出一个进程
            del_pids[p['pid']] = p  # 加入删除
            x -= sum(p['gpu_mem'].values()) / 1024
            n += 1
            error_pids.setdefault(error, set())  # 错误信息
            error_pids[error].add(p['pid'])  # 加入错误pid
        # 单进程
        for p in processes_gpu:
            # 最多显卡(张)
            if len(p['gpu_mem']) > config['gpu_card']:
                del_pids[p['pid']] = p  # 加入删除
                error = '单进程占用的显卡数量 > ' + str(config['gpu_card'])
                error_pids.setdefault(error, set())  # 错误信息
                error_pids[error].add(p['pid'])  # 加入错误pid
            # 最大显存(MB)
            if sum(p['gpu_mem'].values()) > config['gpu_mem']:
                del_pids[p['pid']] = p  # 加入删除
                error = '单进程占用的显存(MB) > ' + str(config['gpu_mem'])
                error_pids.setdefault(error, set())  # 错误信息
                error_pids[error].add(p['pid'])  # 加入错误pid
            # 最长gpu时间(天)
            if p['etime'] / 60 / 24 > config['gpu_day']:
                del_pids[p['pid']] = p  # 加入删除
                error = '单进程占用显卡的时间(天) > ' + str(config['gpu_day'])
                error_pids.setdefault(error, set())  # 错误信息
                error_pids[error].add(p['pid'])  # 加入错误pid
        for p in processes:
            # 最长cpu时间(天)
            if p['etime'] / 60 / 24 > config['cpu_day'] and p['cpu'] > config['cpu_day_core_limit']:
                del_pids[p['pid']] = p  # 加入删除
                error = '单进程占用cpu(%%>%d)的时间(天) > ' % config['cpu_day_core_limit'] + str(config['cpu_day'])
                error_pids.setdefault(error, set())  # 错误信息
                error_pids[error].add(p['pid'])  # 加入错误pid
        # 删除
        if del_pids:
            print(str(datetime.now()), '用户', user, '超出的资源限制/即将杀死的进程PID:')
            pprint(error_pids)
            print('kill-processes:')
            for i, (pid, pro) in enumerate(del_pids.items()):
                print(i + 1, '-', pro)
                if not test:
                    subprocess.getstatusoutput('kill -9 ' + pid)
            print()


def main():
    parser = argparse.ArgumentParser(description='用于限制linux系统的cpu/gpu资源使用')
    parser.add_argument('-c', default='kill.config', help='配置文件路径, 文件不存在会自动创建一个默认配置')  # 配置文件路径
    parser.add_argument('-t', action="store_true", help='是否进行测试, 测试不会杀死进程, 并修改为容易触发kill的参数')
    parser.add_argument('--knp', default='', help='kill nvidia python user, 退出某用户在显卡中的python进程, 会忽略其他功能')
    args, unknown = parser.parse_known_args()  # 忽略未知参数
    if args.knp != '':
        os.system(
            '''up1kEK9m=$(lsof /dev/nvidia* | grep -E python.+%s); echo -e "$up1kEK9m\\nStart killing in 5 seconds"; sleep 5; echo "$up1kEK9m" | awk '{print $2}' | xargs -I {} kill -9 {}''' % args.knp)
        sys.exit(0)
    args.c = os.path.expanduser(args.c)  # 可以使用 ~ 表示用户目录
    if not os.path.exists(args.c):
        # 自动寻找默认 kill.config 的位置
        default_config_path = subprocess.getstatusoutput(
            'py=$(which python) && echo ${py%bin*}lib/python*/site-packages/tsc_auto/kill.config')
        # 没有报错
        if default_config_path[0] == 0:
            default_config_path = default_config_path[1]
        else:
            raise NameError(str(default_config_path) + ' 寻找 ...lib/python*/site-packages/tsc_auto/kill.config 错误!')
        shutil.copyfile(default_config_path, args.c)
        print('创建了一个默认配置:', args.c)
    else:
        print('使用配置:', args.c)
    # 默认配置
    config = {
        's': 10,  # 多少秒检测一次
        'cpu_core_u': 2147483647,  # 一个用户-最多CPU占用(百分比,如100表示占满1个超线程)
        'cpu_mem_u': 2147483647,  # 一个用户-最大内存(GB), 也许是因为进程共享内存, 这个累计内存参数计算不准
        'gpu_card_u': 2,  # 一个用户-最多显卡(张)
        'gpu_mem_u': 30,  # 一个用户-最大显存(GB)
        'gpu_card': 1,  # 单进程-最多显卡(张)
        'gpu_mem': 23000,  # 单进程-最大显存(MB)
        'gpu_day': 20,  # 单进程-最长显卡占用时间(天)
        'cpu_day': 20,  # 单进程-最长cpu占用时间(天)
        'cpu_day_core_limit': 80,  # CPU占用百分比超过此值的进程才会使 cpu_day 配置生效
        'include_u': set(),  # 不可忽略的用户, 优先级高于 ignore_u
        # 忽略的用户, 默认会包含 /etc/passwd 中路径不含有 /home/ 的用户
        'ignore_u': {i.split(':')[0] for i in open('/etc/passwd').readlines() if '/home/' not in i},
        # 'ignore_u': os.popen("cat /etc/passwd | grep -v '/home/' | awk -F ':' '{print $1}'").read().splitlines(),
        # 针对每个特殊配置设置用户，没写的默认使用上述设置，越靠list后面的优先级越高会覆盖前面一样的用户配置
        'conf': [
            {  # 一组配置和对应的用户
                'gpu_mem_u': 45,
                'gpu_card_u': 3,
                'gpu_card': 2,
                'gpu_mem': 25000,
                'conf_u': {'tanshicheng'},  # 使用这组配置的用户
            },
        ],
        # 针对每个用户的额外配置, 没写的默认使用上述设置，优先级最高
        'user': {
            'tanshicheng': {
                'gpu_mem_u': 45,
                'gpu_card_u': 3,
                'gpu_card': 2,
                'gpu_mem': 25000,
            },
        },
    }
    # 加参数测试, 测试的时候使用容易出发kill的参数
    if args.t:
        print('测试...不杀死进程\n')
        config.update({
            'cpu_core_u': 50,
            'cpu_mem_u': 1,
            'gpu_card_u': 0,
            'gpu_mem_u': 0,
            'gpu_card': 0,
            'gpu_mem': 1,
            'gpu_day': 1,
            'cpu_day': 1,
            'cpu_day_core_limit': 0,
        })
        # config['include_u'].add('root')  # 加入 root 进行测试
    # 防止进程重复运行
    if not args.t:
        std = subprocess.Popen(["pgrep", "-f", __file__], stdout=subprocess.PIPE).communicate()
        if len(std[0].decode().split()) > 1:
            exit('Already running')
    # 开始不断检测
    while True:
        if args.t:
            print('\n测试 - ', str(datetime.now()))
        if os.path.isfile(args.c):
            try:
                config_new = copy.deepcopy(config)
                with open(args.c, 'r', encoding='utf8') as r:
                    for k, v in eval(r.read().strip()).items():  # 小心eval恶意插入代码
                        config_new.setdefault(k, v)
                        if isinstance(v, dict):
                            config_new[k].update(v)
                        elif isinstance(v, set):
                            config_new[k] |= v
                        else:
                            config_new[k] = v
                if config != config_new:
                    print(str(datetime.now()), '重新加载约束条件!')  # 没有输出这行注意可能是config路径不对
                    if args.t:
                        pprint(config_new)
                        print()
                    config = config_new
            except:
                traceback.print_exc()
                print()
        kill_processes(config, args.t)
        time.sleep(config['s'])


if __name__ == '__main__':
    main()
